#include <stdlib.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#define LEDS 16
#define PINMASK 31
#define INT_IGNORE 5

unsigned char ledMappings[] PROGMEM = {
	0x12, 0x21,
	0x04, 0x40,

	0x13, 0x31,
	0x30, 0x03,

	0x32, 0x23,
	0x24, 0x42,
	
	0x41, 0x14,
	0x10, 0x01
};

#define FRAMES 45

unsigned int frames[] PROGMEM = {
	0x000f, 0x00f0, 0x0f00, 0xf000, 0x0f00, 0x00f0, 0x000f,
	0x0000,
	0xffff, 0x0000, 
	0x1111, 0x4444, 0x2222, 0x8888,
	0x1111, 0x4444, 0x2222, 0x8888,
	0x1111, 0x4444, 0x2222, 0x8888,
	0x0000,
	0x000f, 0x00ff, 0x0fff, 0xffff, 0xfff0, 0xff00, 0xf000, 0x0000, 0xff00, 0xfff0, 0xffff,
	0x0000,
	0xffff,
	0x0000,
	0xffff,	0xf0f0, 0x0f0f, 0xf0f0, 0x0f0f, 0xf0f0, 0x0f0f, 0xffff,
	0x0000
};

unsigned int displayState;
unsigned char interrupts;
unsigned char frame;

ISR(TIM0_COMPA_vect) {
	interrupts++;
	if(interrupts == INT_IGNORE) {
		interrupts = 0;
//		displayState = rand();
		displayState = pgm_read_dword(&frames[frame++]);
		if(frame == FRAMES) frame = 0;
	}
}

void charlieplex() {
	unsigned char led;
	unsigned char map, set, unset;

	for(led = 0; led < LEDS; led++) {
		map = pgm_read_byte(&ledMappings[led]);
		set = 1 << (map & 0x0f);
		unset = 1 << ((map & 0xf0) >> 4);
		PORTB = 0;
		DDRB = set | unset;
		if(displayState & (1 << led)) PORTB = set;
		_delay_us(200);
	}

	DDRB = 0;
	PORTB = 0;
}

void init() {
	frame = 0;
	displayState = 0;
	
	MCUCR |= _BV(PUD);		// disable pull-ups globally

	TCCR0A = _BV(WGM01);		// Clear timer on compare match mode
	TCCR0B = _BV(CS02) | _BV(CS00);	// Clock prescale by 1024
	OCR0A = 0xff;			// Compare match at top
	TIMSK0 = _BV(OCIE0A);		// Enable compare match interrupts

	sei();
}

int main() {

	init();
	while(1) charlieplex();

	return 0; // unreached
}
